@@ -3,10 +3,23 @@ import os
 
 from six import string_types
 
+from pulsar.managers.util.pykube_util import (
+    ensure_pykube,
+    find_job_object_by_name,
+    find_pod_object_by_name,
+    galaxy_instance_id,
+    Job,
+    job_object_dict,
+    produce_unique_k8s_job_name,
+    pull_policy,
+    pykube_client_from_dict,
+    stop_job,
+)
 from .action_mapper import (
     actions,
     path_type,
 )
+from .amqp_exchange import ACK_FORCE_NOACK_KEY
 from .decorators import parseJson
 from .decorators import retry
 from .destination import submit_params
@@ -21,6 +34,14 @@ from .util import to_base64_json
 log = logging.getLogger(__name__)
 
 CACHE_WAIT_SECONDS = 3
+TOOL_EXECUTION_CONTAINER_COMMAND_TEMPLATE = """
+path='%s/command_line';
+while [ ! -e $path ];
+    do sleep 1; echo "waiting for job script $path";
+done;
+echo 'running script';
+sh $path;
+echo 'ran script'"""
 
 
 class OutputNotFoundException(Exception):
@@ -37,23 +58,12 @@ class BaseJobClient(object):
     def __init__(self, destination_params, job_id):
         destination_params = destination_params or {}
         self.destination_params = destination_params
-        self.job_id = job_id
-        if "jobs_directory" in destination_params:
-            staging_directory = destination_params["jobs_directory"]
-            sep = destination_params.get("remote_sep", os.sep)
-            job_directory = RemoteJobDirectory(
-                remote_staging_directory=staging_directory,
-                remote_id=job_id,
-                remote_sep=sep,
-            )
-        else:
-            job_directory = None
+        self.assign_job_id(job_id)
 
         for attr in ["ssh_key", "ssh_user", "ssh_host", "ssh_port"]:
             setattr(self, attr, destination_params.get(attr, None))
         self.env = destination_params.get("env", [])
         self.files_endpoint = destination_params.get("files_endpoint", None)
-        self.job_directory = job_directory
 
         default_file_action = self.destination_params.get("default_file_action", "transfer")
         if default_file_action not in actions:
@@ -63,6 +73,23 @@ class BaseJobClient(object):
 
         self.setup_handler = build_setup_handler(self, destination_params)
 
+    def assign_job_id(self, job_id):
+        self.job_id = job_id
+        self._set_job_directory()
+
+    def _set_job_directory(self):
+        if "jobs_directory" in self.destination_params:
+            staging_directory = self.destination_params["jobs_directory"]
+            sep = self.destination_params.get("remote_sep", os.sep)
+            job_directory = RemoteJobDirectory(
+                remote_staging_directory=staging_directory,
+                remote_id=self.job_id,
+                remote_sep=sep,
+            )
+        else:
+            job_directory = None
+        self.job_directory = job_directory
+
     def setup(self, tool_id=None, tool_version=None, preserve_galaxy_python_environment=None):
         """
         Setup remote Pulsar server to run this job.
@@ -257,6 +284,10 @@ class JobClient(BaseJobClient):
         }
         self._raw_execute("download_output", output_params, output_path=output_path)
 
+    def job_ip(self):
+        """Return a entry point ports dict (if applicable)."""
+        return None
+
 
 class BaseMessageJobClient(BaseJobClient):
 
@@ -298,6 +329,15 @@ class BaseMessageJobClient(BaseJobClient):
             launch_params["setup_params"] = setup_params
         return launch_params
 
+    def _build_status_request_message(self):
+        # Because this is used to poll, status requests will not be resent if we do not receive an acknowledgement
+        update_params = {
+            'request': 'status',
+            'job_id': self.job_id,
+            ACK_FORCE_NOACK_KEY: True,
+        }
+        return update_params
+
 
 class MessageJobClient(BaseMessageJobClient):
 
@@ -312,7 +352,13 @@ class MessageJobClient(BaseMessageJobClient):
             job_config=job_config,
         )
         response = self.client_manager.exchange.publish("setup", launch_params)
-        log.info("Job published to setup message queue.")
+        log.info("Job published to setup message queue: %s", self.job_id)
+        return response
+
+    def get_status(self):
+        status_params = self._build_status_request_message()
+        response = self.client_manager.exchange.publish("setup", status_params)
+        log.info("Job status request published to setup message queue: %s", self.job_id)
         return response
 
     def kill(self):
@@ -346,6 +392,161 @@ class MessageCLIJobClient(BaseMessageJobClient):
         pass
 
 
+class MessageCoexecutionPodJobClient(BaseMessageJobClient):
+
+    def __init__(self, destination_params, job_id, client_manager):
+        ensure_pykube()
+        super(MessageCoexecutionPodJobClient, self).__init__(destination_params, job_id, client_manager)
+        self.instance_id = galaxy_instance_id(destination_params)
+        self.pulsar_container_image = destination_params.get("pulsar_container_image", "galaxy/pulsar-pod-staging:0.13.0")
+        self._default_pull_policy = pull_policy(destination_params)
+
+    def launch(
+        self,
+        command_line,
+        dependencies_description=None,
+        env=[],
+        remote_staging=[],
+        job_config=None,
+        container_info=None,
+        pulsar_app_config=None
+    ):
+        """
+        """
+        launch_params = self._build_setup_message(
+            command_line,
+            dependencies_description=dependencies_description,
+            env=env,
+            remote_staging=remote_staging,
+            job_config=job_config,
+        )
+        container = None
+        guest_ports = None
+        if container_info is not None:
+            container = container_info.get("container_id")
+            guest_ports = container_info.get("guest_ports")
+
+        manager_type = "coexecution" if container is not None else "unqueued"
+        if "manager" not in pulsar_app_config and "managers" not in pulsar_app_config:
+            pulsar_app_config["manager"] = {"type": manager_type}
+
+        using_dependencies = container is None and dependencies_description is not None
+        if using_dependencies and "dependency_resolution" not in pulsar_app_config:
+            # Setup default dependency resolution for container above...
+            dependency_resolution = {
+                "cache": False,
+                "use": True,
+                "default_base_path": "/pulsar_dependencies",
+                "cache_dir": "/pulsar_dependencies/_cache",
+                "resolvers": [{  # TODO: add CVMFS resolution...
+                    "type": "conda",
+                    "auto_init": True,
+                    "auto_install": True,
+                    "prefix": '/pulsar_dependencies/conda',
+                }, {
+                    "type": "conda",
+                    "auto_init": True,
+                    "auto_install": True,
+                    "prefix": '/pulsar_dependencies/conda',
+                    "versionless": True,
+                }]
+            }
+            pulsar_app_config["dependency_resolution"] = dependency_resolution
+        base64_message = to_base64_json(launch_params)
+        base64_app_conf = to_base64_json(pulsar_app_config)
+
+        job_name = self._k8s_job_name
+        params = self.destination_params
+
+        pulsar_container_image = self.pulsar_container_image
+
+        job_directory = self.job_directory
+
+        volumes = [
+            {"name": "staging-directory", "emptyDir": {}},
+        ]
+        volume_mounts = [
+            {"mountPath": "/pulsar_staging", "name": "staging-directory"},
+        ]
+        pulsar_container_dict = {
+            "name": "pulsar-container",
+            "image": pulsar_container_image,
+            "command": ["pulsar-submit"],
+            "args": ["--base64", base64_message, "--app_conf_base64", base64_app_conf],
+            "workingDir": "/",
+            "volumeMounts": volume_mounts,
+        }
+        tool_container_image = container
+        container_dicts = [pulsar_container_dict]
+        if container:
+            command = TOOL_EXECUTION_CONTAINER_COMMAND_TEMPLATE % job_directory.job_directory
+            tool_container_spec = {
+                "name": "tool-container",
+                "image": tool_container_image,
+                "command": ["sh"],
+                "args": ["-c", command],
+                "workingDir": "/",
+                "volumeMounts": volume_mounts,
+            }
+            if guest_ports:
+                tool_container_spec["ports"] = [{"containerPort": int(p)} for p in guest_ports]
+            container_dicts.append(tool_container_spec)
+        for container_dict in container_dicts:
+            if self._default_pull_policy:
+                container_dict["imagePullPolicy"] = self._default_pull_policy
+
+        template = {
+            "metadata": {
+                "labels": {"app": job_name},
+            },
+            "spec": {
+                "volumes": volumes,
+                "restartPolicy": "Never",
+                "containers": container_dicts,
+            }
+        }
+        spec = {"template": template}
+        k8s_job_obj = job_object_dict(params, job_name, spec)
+        pykube_client = self._pykube_client
+        job = Job(pykube_client, k8s_job_obj)
+        job.create()
+
+    def kill(self):
+        job_name = self._k8s_job_name
+        pykube_client = self._pykube_client
+        job = find_job_object_by_name(pykube_client, job_name)
+        if job:
+            log.info("Kill k8s job with name %s" % job_name)
+            stop_job(job)
+        else:
+            log.info("Attempted to kill k8s job but it is unavailable.")
+
+    def job_ip(self):
+        job_name = self._k8s_job_name
+        pykube_client = self._pykube_client
+        pod = find_pod_object_by_name(pykube_client, job_name)
+        if pod:
+            status = pod.obj['status']
+        else:
+            status = {}
+
+        if 'podIP' in status:
+            pod_ip = status['podIP']
+            return pod_ip
+        else:
+            log.debug("Attempted to get ports dict but k8s pod unavailable")
+
+    @property
+    def _pykube_client(self):
+        return pykube_client_from_dict(self.destination_params)
+
+    @property
+    def _k8s_job_name(self):
+        job_id = self.job_id
+        job_name = produce_unique_k8s_job_name(app_prefix="pulsar", job_id=job_id, instance_id=self.instance_id)
+        return job_name
+
+
 class InputCachingJobClient(JobClient):
     """
     Beta client that cache's staged files to prevent duplication.
